from pyqtgraph.Qt import QtWidgets
from functools import partial


class ParamWindow(QtWidgets.QDialog):
    def __init__(self, spec, parent, filter=[]):
        self.parent = parent
        super().__init__(self.parent)

        self.spec = spec
        self.setWindowTitle(f"Parameters {spec._name}.{'.'.join(spec._depth)}")

        self.layout = QtWidgets.QGridLayout()
        self.labels = []
        self.widgets = []
        row = 0

        for key, value in spec.to_dict().items():
            if key in filter:
                continue
            self.add_widget(key, value, row)
            row += 1
        if row == 0:
            label = QtWidgets.QLabel("No parameters to show.")
            self.layout.addWidget(label, row, 0)
        self.setLayout(self.layout)

    def add_widget(self, key, value, row):
        label = QtWidgets.QLabel(key)
        if type(value) is dict:
            spec = getattr(self.spec, key)

            widget = QtWidgets.QPushButton(f"Inspect {spec._name}.{'.'.join(spec._depth)}")
            widget.pressed.connect(partial(self.open_dict_dialog, key=key))
        else:
            # widget = QtWidgets.QLabel(str(value))
            widget = QtWidgets.QLineEdit(str(value))
            widget.setReadOnly(True)
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        self.layout.addWidget(label, row, 0)
        self.layout.addWidget(widget, row, 1)
        self.labels.append(label)
        self.widgets.append(widget)

    def open_dict_dialog(self, key):
        dialog = ParamWindow(
            spec=getattr(self.spec, key),
            parent=self.parent,
        )
        dialog.exec()



import numpy as np
import networkx as nx
import ast


def tryeval(val):
    try:
        val = ast.literal_eval(val)
    except Exception as e:
        if isinstance(e, ValueError):
            pass
        elif isinstance(e, SyntaxError):
            pass
        else:
            raise
    return val


def empty_gui_state():
    return dict(pos=None, linestyle=dict())


def get_yaml_type(yaml):
    if "node_type" in yaml:
        if "targets" in yaml:
            type = "reset_node"
        else:
            type = "node"
    else:
        type = "object"
    return type


def get_connected_nodes(state, node):
    # Get a set of all the nodes that are connected directly to some node
    connected_inputs = set()
    connected_outputs = set()
    for source, target in state["connects"]:
        if source[0] == node:
            connected_outputs.add(target[0])
        elif target[0] == node:
            connected_inputs.add(source[0])
    return connected_inputs, connected_outputs


def is_connected_to_fixed(state, node, fixed_nodes, nodes_to_ignore=None):
    # Check if node is connected to any fixed node (direct or indirect)
    if nodes_to_ignore is None:
        nodes_to_ignore = set()
    nodes_to_ignore.add(node)
    connected_inputs, connected_outputs = get_connected_nodes(state, node)
    connected_nodes = set.union(connected_inputs, connected_outputs)
    for connected_node in connected_nodes:
        if connected_node in nodes_to_ignore:
            continue
        elif connected_node in fixed_nodes:
            return True
        elif is_connected_to_fixed(state, connected_node, fixed_nodes, nodes_to_ignore):
            return True
    return False


def get_longest_shortest_simple_paths(state, fixed_nodes=None):
    """
    Calculates shortest simple path for each fixed node as source and other node as target.
    Returns the longest for each fixed node.
    """
    if fixed_nodes is None:
        fixed_nodes = {}

    G = nx.DiGraph()
    longest_shortest_simple_paths = {}
    for node in state["nodes"].keys():
        G.add_node(node)
    for source, target in state["connects"]:
        source_name, _, _ = source
        target_name, _, _ = target
        G.add_edge(source_name, target_name)
    for source in fixed_nodes:
        longest_shortest_simple_path = []
        if source not in state["nodes"].keys():
            continue
        for target in state["nodes"].keys():
            if source == target or ("env/observations" in [source, target] and "env/render" in [source, target]):
                continue
            if nx.has_path(G, source=source, target=target):
                simple_paths = nx.all_simple_paths(G, source=source, target=target)
                # Initialize shortest path length with max (i.e. number of nodes) such that each existing path would be shorter
                shortest_path_len = len(state["nodes"].keys())
                shortest_path = []
                for simple_path in simple_paths:
                    if len(simple_path) < shortest_path_len:
                        shortest_path = simple_path
                        shortest_path_len = len(shortest_path)
                if shortest_path_len > len(longest_shortest_simple_path):
                    longest_shortest_simple_path = shortest_path
        longest_shortest_simple_paths[source] = longest_shortest_simple_path
    return longest_shortest_simple_paths


def add_pos_to_state(state, is_engine=False):
    """
    Position nodes using Fruchterman-Reingold force-directed algorithm.
    """
    # Initialize fixed nodes
    left_nodes = {"actuators"} if is_engine else {"env/actions"}
    if is_engine:
        right_nodes = {"sensors"}
    elif "env/render" in state["nodes"].keys():
        right_nodes = {"env/observations", "env/render"}
    else:
        right_nodes = {"env/observations"}

    fixed_nodes = set.union(left_nodes, right_nodes)
    node_size = 150

    # Check for which nodes the position is prescribed
    G = nx.Graph()
    for node, params in state["nodes"].items():
        G.add_node(node)
        if node not in state["gui_state"]:
            state["gui_state"][node] = empty_gui_state()  # Add gui states here.
        gui_state = state["gui_state"][node]
        if gui_state.get("pos", None) is not None and len(gui_state["pos"]) == 2:
            fixed_nodes.add(node)
    for source, target in state["connects"]:
        source_name, _, _ = source
        target_name, _, _ = target
        G.add_edge(source_name, target_name)

    fixed_clusters = []
    loose_clusters = []
    fixed_cluster_nodes = set()
    clusters = nx.connected_components(G)
    for cluster in clusters:
        fixed = False
        for fixed_node in fixed_nodes:
            if fixed_node in cluster:
                fixed_clusters.append(cluster)
                fixed_cluster_nodes.update(cluster)
                fixed = True
                break
        if not fixed:
            loose_clusters.append(cluster)

    # Create Graph
    G = nx.Graph()
    for node in fixed_cluster_nodes:
        G.add_node(node)
    for source, target in state["connects"]:
        source_name, _, _ = source
        target_name, _, _ = target
        if source_name in fixed_cluster_nodes:
            G.add_edge(source_name, target_name)

    # Get longest path in order to set x locations
    longest_shortest_simple_paths = get_longest_shortest_simple_paths(state, fixed_nodes=set.union(left_nodes, right_nodes))

    x_max = 0
    left_max = 0
    right_max = 0
    for left_node in left_nodes:
        len_left = node_size * len(longest_shortest_simple_paths[left_node])
        if len_left > left_max:
            left_max = len_left
        for right_node in right_nodes:
            len_right = node_size * len(longest_shortest_simple_paths[right_node])
            if len_right > right_max:
                right_max = len_right
            if nx.has_path(G, left_node, right_node):
                longest_left = len(longest_shortest_simple_paths[left_node])
                longest_right = len(longest_shortest_simple_paths[right_node])
                max_path_len = node_size * max(longest_left, longest_right)
                if right_node not in longest_shortest_simple_paths[left_node]:
                    max_path_len += node_size
                if max_path_len > x_max:
                    x_max = max_path_len
    if x_max == 0:
        x_max = left_max + right_max + 2 * node_size

    left_connected_clusters = {}
    connected_clusters = {}
    y_pos = {}
    for fixed_node in fixed_nodes:
        connected_inputs, connected_outputs = get_connected_nodes(state, fixed_node)
        if fixed_node in left_nodes:
            left_connected_clusters[fixed_node] = {}
            left_connected_clusters[fixed_node]["in"] = connected_inputs
            left_connected_clusters[fixed_node]["out"] = connected_outputs
        else:
            connected_clusters[fixed_node] = {}
            connected_clusters[fixed_node]["in"] = connected_inputs
            connected_clusters[fixed_node]["out"] = connected_outputs
    for key, connected_cluster in {**left_connected_clusters, **connected_clusters}.items():
        y_pos[key] = {}
        y_pos[key]["in"] = -((len(connected_cluster["in"]) - 1) / 2) * node_size
        y_pos[key]["out"] = -((len(connected_cluster["out"]) - 1) / 2) * node_size

    fixed_positions = dict()
    for fixed_cluster in fixed_clusters:
        for node in sorted(fixed_cluster):
            params = state["nodes"][node]
            gui_state = state["gui_state"][node]
            if params["config"]["name"] in fixed_nodes:
                if gui_state.get("pos", None) is not None and len(gui_state["pos"]) == 2:
                    pos = gui_state["pos"]
                elif params["config"]["name"] in left_nodes:
                    pos = [0, 0]
                elif params["config"]["name"] == "env/render":
                    pos = [x_max, node_size - y_pos["env/observations"]["in"] - y_pos["env/render"]["in"]]
                else:
                    pos = [x_max, 0]
                fixed_positions[node] = pos
    for fixed_cluster in fixed_clusters:
        for node in sorted(fixed_cluster):
            params = state["nodes"][node]
            if not params["config"]["name"] in fixed_nodes:
                for key, clusters in left_connected_clusters.items():
                    for direction in {"in", "out"}:
                        if node in clusters[direction] and node not in fixed_positions.keys():
                            x_pos_cluster = fixed_positions[key][0]
                            if direction == "in":
                                x_pos_cluster -= 1.25 * node_size
                            else:
                                x_pos_cluster += 1.25 * node_size
                            fixed_positions[node] = [x_pos_cluster, fixed_positions[key][1] + y_pos[key][direction]]
                            y_pos[key][direction] += node_size
                            break
                for key, clusters in connected_clusters.items():
                    for direction in {"in", "out"}:
                        if node in clusters[direction] and node not in fixed_positions.keys():
                            x_pos_cluster = fixed_positions[key][0]
                            if direction == "in":
                                x_pos_cluster -= 1.25 * node_size
                            else:
                                x_pos_cluster += 1.25 * node_size
                            fixed_positions[node] = [x_pos_cluster, fixed_positions[key][1] + y_pos[key][direction]]
                            y_pos[key][direction] += node_size
                            break

    position_dict = nx.spring_layout(G, k=1, pos=fixed_positions, fixed=fixed_positions.keys())
    for node, pos in position_dict.items():
        state["gui_state"][node]["pos"] = pos.tolist()
    y_offset = np.max(np.array(list(position_dict.values()))[:, 1]) + node_size
    x_offset = x_max // 2
    for loose_cluster in loose_clusters:
        G = nx.Graph()
        for node in sorted(loose_cluster):
            G.add_node(node)
        for source, target in state["connects"]:
            source_name, _, _ = source
            target_name, _, _ = target
            if source_name in loose_cluster:
                G.add_edge(source_name, target_name)
        position_dict = nx.spring_layout(G, k=1)
        height = (
            abs(
                np.max(np.array(list(position_dict.values()))[:, 1]) + np.min(np.array(list(position_dict.values()))[:, 1]) + 2
            )
            * node_size
        )
        y_offset += height // 2
        if np.min(np.array(list(position_dict.values()))[:, 1]) * node_size < 0:
            y_offset += abs(np.min(np.array(list(position_dict.values()))[:, 1]) * node_size)
        for node, pos in position_dict.items():
            pos *= node_size
            pos += np.array([x_offset, y_offset])
            state["gui_state"][node]["pos"] = pos.tolist()
        y_offset += height // 2
    return state



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Terminal.py
"""

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph import functions as fn
from pyqtgraph.Point import Point

from eagerx_gui import configuration
from eagerx_gui.pyqtgraph_utils import ParamWindow


class GuiTerminal(object):
    def __init__(self, node, name, pos=None):
        self.node = node
        self.name = name
        self.pos = pos
        self.optional = False
        self.connections = {}
        self.node_type = node.node_type

        name_split = name.split("/")
        self.terminal_type = name_split[0]
        self.terminal_name = name_split[-1]

        assert self.terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT), (
            f"Invalid terminal type: {self.terminal_type}, "
            f"should be one of {set.union(configuration.TERMS_IN, configuration.TERMS_OUT)}"
        )

        self.is_input = self.terminal_type in configuration.TERMS_IN
        self.is_state = self.terminal_type in ["targets", "states"]
        self.is_feedthrough = self.terminal_type == "feedthroughs"
        self.is_removable = False
        self.is_renamable = False
        self.is_connectable = False

        # todo: Check if we want to show different parameters if feedthrough (e.g. of the corresponding output)
        self.spec = getattr(getattr(node.gui.graph.get_spec(node.name), self.terminal_type), self.terminal_name)

        self._graphicsItem = TerminalGraphicsItem(self, parent=self.node.graphics_item())
        self.recolor()

    def connection_view(self):
        return self.node.gui.tuple_to_view(self.connection_tuple())

    def connection_tuple(self):
        return (self.node.name, self.terminal_type, self.terminal_name)

    def connected(self, term):
        """Called whenever this terminal has been connected to another.
        (note--this function is called on both terminals)"""
        self.node.connected(self, term)

    def params(self):
        return self.spec.to_dict()

    def is_connected(self):
        return len(self.connections) > 0

    def graphics_item(self):
        return self._graphicsItem

    def connected_to(self, term):
        return term in self.connections

    def has_input(self):
        for t in self.connections:
            if not t.is_input:
                return True
        return False

    def input_terminals(self):
        """Return the terminal(s) that give input to this one."""
        return [t for t in self.connections if not t.is_input]

    def dependent_nodes(self):
        """Return the list of nodes which receive input from this terminal."""
        return set([t.node for t in self.connections if t.is_input])

    def disconnect_all(self):
        for t in list(self.connections.keys()):
            self.disconnect_from(t)

    def recolor(self, color=None, recurse=True):
        if color is None:
            if self.is_state:
                if self.is_connectable:
                    color = QtGui.QColor(255, 0, 0)
                else:
                    color = QtGui.QColor(128, 128, 128)
            elif self.is_feedthrough:
                color = QtGui.QColor(175, 238, 238)
            else:
                color = QtGui.QColor(0, 0, 255)
        self.graphics_item().setBrush(QtGui.QBrush(color))

        if recurse:
            for t in self.connections:
                t.recolor(color, recurse=False)

    def __repr__(self):
        return "<Terminal %s.%s>" % (str(self.node.name), str(self.terminal_name))

    def __hash__(self):
        return id(self)


class TextItem(QtWidgets.QGraphicsTextItem):
    def __init__(self, text, parent, on_update):
        super().__init__(text, parent)
        self.on_update = on_update

    def focusOutEvent(self, ev):
        super().focusOutEvent(ev)
        if self.on_update is not None:
            self.on_update()

    def keyPressEvent(self, ev):
        super().keyPressEvent(ev)


class TerminalGraphicsItem(GraphicsObject):
    def __init__(self, term, parent=None):
        self.term = term
        GraphicsObject.__init__(self, parent)
        self.brush = fn.mkBrush(0, 0, 0)
        self.box = QtWidgets.QGraphicsRectItem(0, 0, 10, 10, self)
        on_update = None
        self.label = TextItem(self.term.terminal_name, self, on_update)
        self.label.setScale(0.7)
        self.newConnection = None
        self.setFiltersChildEvents(True)  # to pick up mouse events on the rectitem
        self.setZValue(1)
        self.menu = None

    def label_focus_out(self, ev):
        QtWidgets.QGraphicsTextItem.focusOutEvent(self.label, ev)
        self.label_changed()

    def label_key_press(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            self.label_changed()
        else:
            QtWidgets.QGraphicsTextItem.keyPressEvent(self.label, ev)

    def setBrush(self, brush):
        self.brush = brush
        self.box.setBrush(brush)

    def boundingRect(self):
        br = self.box.mapRectToParent(self.box.boundingRect())
        lr = self.label.mapRectToParent(self.label.boundingRect())
        return br | lr

    def paint(self, p, *args):
        pass

    def set_anchor(self, x, y):
        pos = QtCore.QPointF(x, y)
        self.anchorPos = pos
        br = self.box.mapRectToParent(self.box.boundingRect())
        lr = self.label.mapRectToParent(self.label.boundingRect())

        if self.term.is_input:
            self.box.setPos(pos.x(), pos.y() - br.height() / 2.0)
            self.label.setPos(pos.x() + br.width(), pos.y() - lr.height() / 2.0)
        else:
            self.box.setPos(pos.x() - br.width(), pos.y() - br.height() / 2.0)
            self.label.setPos(pos.x() - br.width() - lr.width(), pos.y() - lr.height() / 2.0)
        self.update_connections()

    def update_connections(self):
        for t, c in self.term.connections.items():
            c.update_line()

    def mousePressEvent(self, ev):
        ev.ignore()  # necessary to allow click/drag events to process correctly

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            self.label.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)
        elif ev.button() == QtCore.Qt.MouseButton.RightButton:
            ev.accept()
            self.raise_context_menu(ev)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            # todo: find out which keys to suppress in dialog
            filter = []
            terminal_type = self.term.terminal_type
            widgets_to_hide = configuration.GUI_WIDGETS["term"]["hide"]
            for key in self.term.params().keys():
                if key in widgets_to_hide["all"]:
                    filter.append(key)
                elif terminal_type in widgets_to_hide and key in widgets_to_hide[terminal_type]:
                    filter.append(key)
            param_window = ParamWindow(spec=self.term.spec, parent=self.parent(), filter=filter)
            param_window.exec()
            param_window.close()

    def raise_context_menu(self, ev):
        # only raise menu if this terminal is removable
        ev.ignore()

    def mouseDragEvent(self, ev):
        return

    def hoverEvent(self, ev):
        if not ev.isExit() and ev.acceptDrags(QtCore.Qt.MouseButton.LeftButton):
            ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton)
            # we don't use the click, but we also don't want anyone else to use it.
            ev.acceptClicks(QtCore.Qt.MouseButton.RightButton)
            self.box.setBrush(fn.mkBrush("w"))
        else:
            self.box.setBrush(self.brush)
        self.update()

    def connect_point(self):
        # return the connect position of this terminal in view coords
        return self.mapToView(self.mapFromItem(self.box, self.box.boundingRect().center()))

    def node_moved(self):
        for t, item in self.term.connections.items():
            item.update_line()


class ConnectionItem(GraphicsObject):
    def __init__(self, source, target=None):
        GraphicsObject.__init__(self)
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsFocusable)
        self.source = source
        self.target = target
        self.length = 0
        self.hovered = False
        self.path = None
        self.shapePath = None
        self.connection_window = None

        if self.source.term.is_state:
            self.style = {
                "shape": "line",
                "color": (255, 0, 0, 255),
                "width": 2.0,
                "hoverColor": (150, 150, 250, 255),
                "hoverWidth": 2.0,
                "selectedColor": (200, 200, 0, 255),
                "selectedWidth": 4.0,
            }
        else:
            self.style = {
                "shape": "line",
                "color": (0, 0, 255, 255),
                "width": 2.0,
                "hoverColor": (150, 150, 250),
                "hoverWidth": 2.0,
                "selectedColor": (200, 200, 0, 255),
                "selectedWidth": 4.0,
            }
        self.source.getViewBox().addItem(self)
        self.update_line()

    def set_target(self, target):
        self.target = target
        self.update_line()

    def setStyle(self, **kwds):
        self.style.update(kwds)
        if "shape" in kwds:
            self.update_line()
        else:
            self.update()

    def update_line(self):
        start = Point(self.source.connect_point())
        if isinstance(self.target, TerminalGraphicsItem):
            stop = Point(self.target.connect_point())
            input_term = self.target.term if self.target.term.is_input else self.source.term
            node_name = input_term.node.name
            linestyle_state = input_term.node.gui.graph._state["gui_state"][node_name]["linestyle"]
            if input_term.name in linestyle_state:
                shape = linestyle_state[input_term.name]
            else:
                shape = "cubic"
                linestyle_state[input_term.name] = shape
            self.style["shape"] = shape
        elif isinstance(self.target, QtCore.QPointF):
            stop = Point(self.target)
        else:
            return
        self.prepareGeometryChange()

        self.path = self.generate_path(start, stop)
        self.shapePath = None
        self.update()
        self.setZValue(-1)

    def generate_path(self, start, stop):
        path = QtGui.QPainterPath()
        path.moveTo(start)
        if self.style["shape"] == "line":
            path.lineTo(stop)
        elif self.style["shape"] == "cubic":
            path.cubicTo(
                Point(stop.x(), start.y()),
                Point(start.x(), stop.y()),
                Point(stop.x(), stop.y()),
            )
        else:
            raise Exception('Invalid shape "%s"; options are "line" or "cubic"' % self.style["shape"])
        return path

    def keyPressEvent(self, ev):
        if not self.isSelected():
            ev.ignore()
            return

        if ev.key() == QtCore.Qt.Key.Key_Control:
            ev.accept()
            if self.style["shape"] == "line":
                shape = "cubic"
            elif self.style["shape"] == "cubic":
                shape = "line"
            if isinstance(self.target, TerminalGraphicsItem):
                # Update linestyle in gui state
                input_term = self.target.term if self.target.term.is_input else self.source.term
                node_name = input_term.node.name
                input_term.node.gui.graph._state["gui_state"][node_name]["linestyle"][input_term.name] = shape
                self.setStyle(shape=shape)
        else:
            ev.ignore()

    def mousePressEvent(self, ev):
        ev.ignore()

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            sel = self.isSelected()
            self.setSelected(True)
            self.setFocus()
            if not sel and self.isSelected():
                self.update()

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            if self.connection_window is not None:
                self.connection_window.show()

    def hoverEvent(self, ev):
        if (not ev.isExit()) and ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton):
            self.hovered = True
        else:
            self.hovered = False
        self.update()

    def boundingRect(self):
        return self.shape().boundingRect()

    def viewRangeChanged(self):
        self.shapePath = None
        self.prepareGeometryChange()

    def shape(self):
        if self.shapePath is None:
            if self.path is None:
                return QtGui.QPainterPath()
            stroker = QtGui.QPainterPathStroker()
            px = self.pixelWidth()
            stroker.setWidth(px * 8)
            self.shapePath = stroker.createStroke(self.path)
        return self.shapePath

    def paint(self, p, *args):
        if self.isSelected():
            p.setPen(fn.mkPen(self.style["selectedColor"], width=self.style["selectedWidth"]))
        else:
            if self.hovered:
                p.setPen(fn.mkPen(self.style["hoverColor"], width=self.style["hoverWidth"]))
            else:
                p.setPen(fn.mkPen(self.style["color"], width=self.style["width"]))

        p.drawPath(self.path)



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Flowchart.py
"""

# Import pyqtgraph modules
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph.Qt import QtCore, QtWidgets
from pyqtgraph import dockarea
from pyqtgraph.debug import printExc

# eagerx imports
from eagerx.core.graph import Graph
from eagerx.core.graph_engine import EngineGraph

# eagerx_gui imports
from eagerx_gui import configuration
from eagerx_gui.utils import add_pos_to_state
from eagerx_gui import gui_view
from eagerx_gui.gui_node import RxGuiNode, NodeGraphicsItem
from eagerx_gui.gui_terminal import TerminalGraphicsItem, ConnectionItem


class Gui(QtCore.QObject):
    def __init__(self, state, is_engine=False):
        super().__init__()
        self.graph = Graph(state=state) if not is_engine else EngineGraph(state=state)
        self.is_engine = is_engine
        self.nodes = {}
        self.next_z_val = 10
        self._widget = None
        self.scene = None
        self.file_path = None
        self.widget()
        self.load_state()
        self.viewBox.autoRange(padding=0.04)

    def create_node(self, name, pos):
        """Create a new Node and add it to this flowchart."""
        node = RxGuiNode(name, gui=self)
        self.add_node(node, name, pos)
        return node

    def add_node(self, node, name, pos=None):
        """Add an existing Node to this flowchart.

        See also: createnode
        """
        if pos is None:
            pos = [0, 0]
        if type(pos) in [QtCore.QPoint, QtCore.QPointF]:
            pos = [pos.x(), pos.y()]
        item = node.graphics_item()
        item.setZValue(self.next_z_val * 2)
        self.next_z_val += 1
        self.viewBox.addItem(item)
        item.moveBy(*pos)
        self.nodes[name] = node

    def connect_terminals(self, term1, term2):
        """Connect two terminals together within this flowchart."""
        connection_item = ConnectionItem(term1.graphics_item(), term2.graphics_item())
        term1.graphics_item().getViewBox().addItem(connection_item)

        term1.connections[term2] = connection_item
        term2.connections[term1] = connection_item

        term1.connected(term2)
        term2.connected(term1)

    def chart_graphics_item(self):
        """Return the graphicsItem that displays the internal nodes and
        connections of this flowchart.

        Note that the similar method `graphicsItem()` is inherited from Node
        and returns the *external* graphical representation of this flowchart."""
        return self.viewBox

    def widget(self):
        """Return the control widget for this flowchart.

        This widget provides GUI access to the parameters for each node and a
        graphical representation of the flowchart.
        """
        if self._widget is None:
            self._widget = EagerxGraphWidget(self)
            self.scene = self._widget.scene
            self.viewBox = self._widget.viewBox()
        return self._widget

    def load_state(self):
        self.blockSignals(True)
        try:
            add_pos_to_state(self.graph._state, is_engine=self.is_engine)
            nodes = self.graph._state["nodes"]
            nodes = [dict(**node, name=name, gui_state=self.graph._state["gui_state"][name]) for name, node in nodes.items()]
            nodes.sort(key=lambda a: a["gui_state"]["pos"][0])
            for n in nodes:
                if n["name"] in self.nodes:
                    self.nodes[n["name"]].load_state(n)
                    continue
                try:
                    name = n["name"]
                    pos = n["gui_state"]["pos"]
                    node = self.create_node(name, pos)
                    node.load_state(n)
                except Exception:
                    printExc("Error creating node %s: (continuing anyway)" % n["name"])

            connects = [
                (
                    connection[0][0],
                    "/".join(connection[0][1:3]),
                    connection[1][0],
                    "/".join(connection[1][1:3]),
                )
                for connection in self.graph._state["connects"]
            ]
            for n1, t1, n2, t2 in connects:
                try:
                    self.connect_terminals(self.nodes[n1].terminals[t1], self.nodes[n2].terminals[t2])
                except Exception:
                    print(self.nodes[n1].terminals)
                    print(self.nodes[n2].terminals)
                    printExc("Error connecting terminals %s.%s - %s.%s:" % (n1, t1, n2, t2))

        finally:
            self.blockSignals(False)

    def tuple_to_view(self, t):
        if isinstance(t, tuple):
            t = list(t)
        return self.graph.get_view(t[0], depth=t[1:])


class GraphicsItem(GraphicsObject):
    def __init__(self, chart):
        GraphicsObject.__init__(self)
        self.chart = chart  # chart is an instance of Flowchart()
        self.update_terminals()

    def update_terminals(self):
        self.terminals = {}

    def boundingRect(self):
        return QtCore.QRectF()

    def paint(self, p, *args):
        pass


class EagerxGraphWidget(dockarea.DockArea):
    """Includes the actual graphical flowchart and debugging interface"""

    def __init__(self, chart):
        dockarea.DockArea.__init__(self)
        self.chart = chart
        self.cwWin = QtWidgets.QMainWindow()
        self.cwWin.setWindowTitle("EAGERx EngineGraph")
        self.cwWin.setCentralWidget(self)
        self.cwWin.resize(1000, 800)
        self.hoverItem = None

        # build user interface (it was easier to do it here than via developer)
        self.view = gui_view.RxView(self)
        self.viewDock = dockarea.Dock("view", size=(1000, 600))
        self.viewDock.addWidget(self.view)
        self.viewDock.hideTitleBar()
        self.addDock(self.viewDock)

        self.hoverText = QtWidgets.QTextEdit()
        self.hoverText.setReadOnly(True)
        self.hoverDock = dockarea.Dock("Hover Info", size=(1000, 20))
        self.hoverDock.addWidget(self.hoverText)
        self.addDock(self.hoverDock, "bottom")

        self.scene = self.view.scene()
        self._viewBox = self.view.viewBox()
        self.scene.sigMouseHover.connect(self.hover_over)

    def menuPosChanged(self, pos):
        self.menuPos = pos

    def showViewMenu(self, ev):
        self.buildMenu(ev.scenePos())
        self.nodeMenu.popup(ev.screenPos())

    def viewBox(self):
        return self._viewBox  # the viewBox that items should be added to

    def hover_over(self, items):
        for item in items:
            self.hoverItem = item
            if hasattr(item, "term") and isinstance(item, TerminalGraphicsItem):
                text = "name: " + item.term.terminal_name
                for key, value in item.term.params().items():
                    if key in configuration.GUI_WIDGETS["term"]["hide"]["all"]:
                        continue
                    elif (
                        item.term.node_type in configuration.GUI_WIDGETS["term"]["hide"]
                        and key in configuration.GUI_WIDGETS["term"]["hide"][item.term.node_type]
                    ):
                        continue
                    elif (
                        item.term.terminal_type in configuration.GUI_WIDGETS["term"]["hide"]
                        and key in configuration.GUI_WIDGETS["term"]["hide"][item.term.terminal_type]
                    ):
                        continue
                    text += "\n" + "{}: {}".format(key, value)
                self.hoverText.setPlainText(text)
                return
            elif hasattr(item, "node") and isinstance(item, NodeGraphicsItem):
                text = ""
                for key, value in item.node.params().items():
                    if key in configuration.GUI_WIDGETS["node"]["hide"]["all"]:
                        continue
                    elif (
                        item.node.node_type in configuration.GUI_WIDGETS["node"]["hide"]
                        and key in configuration.GUI_WIDGETS["node"]["hide"][item.node.node_type]
                    ):
                        continue
                    text += "{}: {}\n".format(key, value)
                self.hoverText.setPlainText(text)
                return
        self.hoverText.setPlainText("")



from copy import deepcopy

# Terminal types
TERMS = {
    "object": {"in": {"actuators"}, "out": {"sensors", "states"}},
    "node": {"in": {"inputs"}, "out": {"outputs", "states"}},
    "reset_node": {
        "in": {"feedthroughs", "inputs", "targets"},
        "out": {"outputs", "states"},
    },
}
TERMS_IN = set().union(*[TERMS[key]["in"] for key in TERMS])
TERMS_OUT = set().union(*[TERMS[key]["out"] for key in TERMS])

# Possible entries in GUI
GUI_WIDGETS = {
    "node": {
        "hide": {
            "all": [],
            "actions": ["rate"],
            "observations": ["rate"],
        },
    },
    "term": {
        "hide": {
            "all": ["address"],
            "states": ["processor"],
            "observations": ["sync", "rate"],
            "outputs": ["rate"],
            "inputs": ["rate"],
        },
    },
}

# Corresponding RGB values for colors
GUI_COLORS = {
    "black": [0, 0, 0],
    "red": [255, 0, 0],
    "green": [0, 128, 0],
    "yellow": [255, 255, 0],
    "blue": [0, 0, 255],
    "magenta": [255, 0, 255],
    "cyan": [0, 255, 255],
    "white": [255, 255, 255],
}



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Node.py
"""

import numpy as np

from eagerx_gui import configuration
from eagerx_gui.utils import get_yaml_type
from eagerx_gui.gui_terminal import GuiTerminal
from eagerx_gui.pyqtgraph_utils import ParamWindow

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph import functions as fn
from collections import OrderedDict

translate = QtCore.QCoreApplication.translate


class RxGuiNode(QtCore.QObject):
    sigTerminalAdded = QtCore.Signal(object, object)  # self, term

    def __init__(self, name, gui):
        QtCore.QObject.__init__(self)
        self.name = name
        self._graphics_item = None
        self.terminals = OrderedDict()
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.gui = gui
        self.spec = self.gui.graph.get_spec(name).config

        if self.gui.is_engine:
            self.is_object = False
            self.node_type = "engine_node"
        elif name in ["env/actions", "env/observations", "env/render"]:
            # todo: check if sensors actuators should be added here
            self.node_type = name.split("/")[-1]
            self.is_object = False
        else:
            self.node_type = get_yaml_type(self.gui.graph.get_spec(name).params)
        self.allow_remove = True
        self.allow_add_terminal = False
        self.__initialize_terminals()

    def __next_terminal_name(self, name):
        """Return an unused terminal name"""
        name2 = name
        i = 1
        while name2 in self.terminals:
            name2 = "%s_%d" % (name, i)
            i += 1
        return name2

    def __initialize_terminals(self):
        for terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT):
            if self.node_type == "render" and terminal_type == "outputs":
                continue
            if terminal_type in self.params():
                for terminal in self.params()[terminal_type]:
                    # Filter out terminals step, and set for actions, observations and render nodes
                    if (
                        (self.node_type == "actions" and (terminal_type == "inputs" or terminal == "set"))
                        or (self.node_type in ["observations", "render"] and terminal_type == "outputs")
                        or (self.node_type == "observations" and terminal == "actions_set")
                    ):
                        continue
                    name = terminal_type + "/" + terminal
                    self.add_terminal(name=name)
                    if self.node_type == "reset_node" and terminal_type == "outputs":
                        name = "feedthroughs/" + terminal
                        self.add_terminal(name=name)

    def params(self):
        return self.spec.to_dict()

    def get_view(self):
        return self.gui.get_view(self.name, depth=["config"])

    def add_terminal(self, name):
        """Add a new terminal to this Node with the given name.

        Causes sigTerminalAdded to be emitted."""
        name = self.__next_terminal_name(name)

        term = GuiTerminal(self, name)
        self.terminals[name] = term

        if term.is_input:
            self.inputs[name] = term
        else:
            self.outputs[name] = term

        self.graphics_item().update_terminals()
        self.sigTerminalAdded.emit(self, term)
        return term

    def graphics_item(self):
        """Return the GraphicsItem for this node."""
        if self._graphics_item is None:
            self._graphics_item = NodeGraphicsItem(self)
        return self._graphics_item

    def dependent_nodes(self):
        """Return the list of nodes which provide direct input to this node"""
        nodes = set()
        for t in self.inputs.values():
            nodes |= set([i.node for i in t.input_terminals()])
        return nodes

    def __repr__(self):
        return "<Node %s @%x>" % (self.name, id(self))

    def connected(self, local_term, remote_term):
        """Called whenever one of this node's terminals is connected elsewhere."""
        pass

    def recolor(self):
        self.graphics_item().setPen(QtGui.QPen(QtGui.QColor(150, 0, 0), 3))

    def load_state(self, state):
        pos = state["gui_state"].get("pos", (0, 0))
        self.graphics_item().setPos(*pos)


class TextItem(QtWidgets.QGraphicsTextItem):
    def __init__(self, text, parent, on_update):
        super().__init__(text, parent)
        self.on_update = on_update

    def focusOutEvent(self, ev):
        super().focusOutEvent(ev)
        if self.on_update is not None:
            self.on_update()

    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            if self.on_update is not None:
                self.on_update()
                return
        super().keyPressEvent(ev)

    def mousePressEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            self.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)  # focus text label


class NodeGraphicsItem(GraphicsObject):
    def __init__(self, node):
        GraphicsObject.__init__(self)
        self.node = node
        self._node_type = node.node_type
        flags = (
            self.GraphicsItemFlag.ItemIsMovable
            | self.GraphicsItemFlag.ItemIsSelectable
            | self.GraphicsItemFlag.ItemIsFocusable
            | self.GraphicsItemFlag.ItemSendsGeometryChanges
        )

        self.set_color()
        self.setFlags(flags)
        self.bounds = QtCore.QRectF(0, 0, 125, 125)
        self.nameItem = TextItem(self.node.name, self, self.label_changed)
        self.nameItem.setDefaultTextColor(QtGui.QColor(50, 50, 50))
        self.nameItem.moveBy(self.bounds.width() / 2.0 - self.nameItem.boundingRect().width() / 2.0, 0)
        self.update_terminals()

        self.nameItem.focusOutEvent = self.label_focus_out
        self.nameItem.keyPressEvent = self.label_key_press

        self.menu = None
        self.buildMenu()
        self.initial_z_value = self.zValue()
        self.label_changed()

    def set_color(self):
        if "color" in self.node.params() and self.node.params()["color"] in configuration.GUI_COLORS:
            brush_color = np.array(configuration.GUI_COLORS[self.node.params()["color"]])
        else:
            brush_color = np.array([200, 200, 200])

        self.brush = fn.mkBrush(*brush_color, 50)
        self.hoverBrush = fn.mkBrush(*brush_color, 100)
        self.selectBrush = fn.mkBrush(*brush_color, 100)
        self.pen = fn.mkPen(0, 0, 0, 200, width=2)
        self.selectPen = fn.mkPen(0, 0, 0, 200, width=4)
        self.hovered = False
        self.update()

    def label_focus_out(self, ev):
        QtWidgets.QGraphicsTextItem.focusOutEvent(self.nameItem, ev)
        self.label_changed()

    def label_key_press(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            self.label_changed()
        else:
            QtWidgets.QGraphicsTextItem.keyPressEvent(self.nameItem, ev)

    def label_changed(self):
        new_name = str(self.nameItem.toPlainText())
        if self.node.gui is not None and new_name in self.node.gui.nodes.keys():
            self.nameItem.setPlainText(self.node.name)
            return
        if new_name != self.node.name:
            self.node.rename(new_name)
        self.update_terminals()

    def setPen(self, *args, **kwargs):
        self.pen = fn.mkPen(*args, **kwargs)
        self.update()

    def setBrush(self, brush):
        self.brush = brush
        self.update()

    def update_terminals(self):
        # re-center the label
        if self.nameItem.boundingRect().width() > self.bounds.width() - 5.0:
            self.bounds = QtCore.QRectF(0, 0, self.nameItem.boundingRect().width() + 5.0, self.bounds.height())
        bounds = self.boundingRect()
        self.nameItem.setPos(bounds.width() / 2.0 - self.nameItem.boundingRect().width() / 2.0, 0)

        # Update bounds
        bounds = self.bounds
        self.terminals = {}
        inp = self.node.inputs
        y = self.nameItem.boundingRect().height() + 5
        max_width_inp = 0.0
        for i in sorted(inp.keys()):
            t = inp[i]
            item = t.graphics_item()
            item.setParentItem(self)
            br = self.bounds
            item.set_anchor(0, y)
            self.terminals[i] = (t, item)
            y += item.boundingRect().height() + 2
            if item.boundingRect().width() > max_width_inp:
                max_width_inp = item.boundingRect().width()
        y_inp = y

        out = self.node.outputs
        dy = bounds.height() / (len(out) + 1)
        y = self.nameItem.boundingRect().height() + 5
        max_width_out = 0.0
        for i in sorted(out.keys()):
            t = out[i]
            item = t.graphics_item()
            item.setParentItem(self)
            item.setZValue(self.initial_z_value)
            br = self.bounds
            item.set_anchor(bounds.width(), y)
            self.terminals[i] = (t, item)
            y += item.boundingRect().height() + 2
            if item.boundingRect().width() > max_width_out:
                max_width_out = item.boundingRect().width()
        y_max = max(y_inp, y)

        if y_max > bounds.height() - 5:
            self.bounds = QtCore.QRectF(0, 0, self.bounds.width(), y_max + 6.0)

        if max_width_inp + max_width_out > bounds.width() - 10:
            self.bounds = QtCore.QRectF(0, 0, max_width_inp + max_width_out + 11.0, self.bounds.height())
            self.update_terminals()
        self.update()

    def boundingRect(self):
        return self.bounds.adjusted(-5, -5, 5, 5)

    def paint(self, p, *args):
        p.setPen(self.pen)
        if self.isSelected():
            p.setPen(self.selectPen)
            p.setBrush(self.selectBrush)
            self.setZValue(200)
        else:
            p.setPen(self.pen)
            if self.hovered:
                p.setBrush(self.hoverBrush)
                self.setZValue(200)
            else:
                p.setBrush(self.brush)
                self.setZValue(self.initial_z_value)

        p.drawRect(self.bounds)

    def mousePressEvent(self, ev):
        ev.ignore()

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            sel = self.isSelected()
            self.setSelected(True)
            if not sel and self.isSelected():
                self.update()

        elif ev.button() == QtCore.Qt.MouseButton.RightButton:
            self.menu = None
            self.buildMenu()
            ev.accept()
            self.raiseContextMenu(ev)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            # todo: find out which keys to suppress in dialog
            filter = []
            widgets_to_hide = configuration.GUI_WIDGETS["node"]["hide"]
            for key in self.node.params().keys():
                if key in widgets_to_hide["all"]:
                    filter.append(key)
                elif self._node_type in widgets_to_hide and key in widgets_to_hide[self._node_type]:
                    filter.append(key)
            param_window = ParamWindow(spec=self.node.spec, parent=self.parent(), filter=filter)
            param_window.exec()
            param_window.close()

    def mouseDragEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            self.setPos(self.pos() + self.mapToParent(ev.pos()) - self.mapToParent(ev.lastPos()))

    def hoverEvent(self, ev):
        if not ev.isExit() and ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton):
            ev.acceptDrags(QtCore.Qt.MouseButton.LeftButton)
            self.hovered = True
        else:
            self.hovered = False
        self.update()

    def keyPressEvent(self, ev):
        ev.ignore()

    def itemChange(self, change, val):
        if change == self.GraphicsItemChange.ItemPositionHasChanged:
            self.node.gui.graph._state["gui_state"][self.node.name]["pos"] = [
                self.pos().x(),
                self.pos().y(),
            ]
            for k, t in self.terminals.items():
                t[1].node_moved()
        return GraphicsObject.itemChange(self, change, val)

    def getMenu(self):
        return self.menu

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))

    def buildMenu(self):
        self.menu = QtWidgets.QMenu()
        self.menu.setTitle("Node")



__version__ = "0.2.10"

import os

# Set PyQt backend
for lib in ["PyQt6", "PyQt5"]:
    try:
        __import__(lib)
        os.environ["PYQTGRAPH_QT_LIB"] = lib
        break
    except ImportError:
        pass

import pyqtgraph as pg
import pyqtgraph.exporters
import sys
import numpy as np
from pyqtgraph.Qt import QtWidgets, QtGui
from eagerx_gui.gui import Gui


def launch_gui(state, is_engine=False):
    app = QtWidgets.QApplication(sys.argv)

    ## Create main window with grid layout
    win = QtWidgets.QMainWindow()
    win.setWindowTitle("EAGERx Graph")
    cw = QtWidgets.QWidget()
    win.setCentralWidget(cw)
    layout = QtWidgets.QGridLayout()
    cw.setLayout(layout)

    rx_gui = Gui(state, is_engine=is_engine)
    w = rx_gui.widget()

    # Add flowchart control panel to the main window
    layout.addWidget(w, 0, 0, 2, 1)

    win.show()

    app.exec()
    state["gui_state"] = rx_gui.graph._state["gui_state"]
    app.quit()
    return state


def render_gui(state, resolution=None, is_engine=False, filename=None):
    if resolution is None:
        resolution = [1920, 1080]

    app = QtWidgets.QApplication(sys.argv)

    ## Create main window with grid layout
    win = QtWidgets.QMainWindow()
    win.setWindowTitle("EAGERx Graph")
    cw = QtWidgets.QWidget()
    win.setCentralWidget(cw)
    layout = QtWidgets.QGridLayout()
    cw.setLayout(layout)

    rx_gui = Gui(state, is_engine=is_engine)
    w = rx_gui.widget()

    # Add flowchart control panel to the main window
    layout.addWidget(w, 0, 0, 2, 1)

    win.show()

    if filename is not None:
        if not filename.endswith(".svg"):
            filename += ".svg"
        svgexporter = pg.exporters.SVGExporter(w.view.scene())
        svgexporter.export(filename)

    exporter = pg.exporters.ImageExporter(w.view.scene())
    exporter.parameters()["width"] = resolution[0]
    exporter.parameters()["height"] = resolution[1]

    png = exporter.export(toBytes=True)
    png.convertToFormat(QtGui.QImage.Format.Format_RGB32)

    width = png.width()
    height = png.height()

    ptr = png.bits()
    ptr.setsize(height * width * 4)
    arr = np.array(ptr).reshape(height, width, 4)  # Copies the data
    app.quit()
    return arr[..., [2, 1, 0, 3]]


def launch_engine_gui(state):
    return launch_gui(state, is_engine=True)


def render_engine_gui(state):
    return render_gui(state, is_engine=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagerx_gui']

package_data = \
{'': ['*']}

install_requires = \
['PyQt6>=6.2.3,<7.0.0',
 'PyVirtualDisplay>=3.0,<4.0',
 'eagerx>=0.1.30,<0.2.0',
 'opencv-python>=4.5,<5.0',
 'pyqtgraph>=0.12.4,<0.13.0']

setup_kwargs = {
    'name': 'eagerx-gui',
    'version': '0.2.10',
    'description': 'GUI to visualise graphs in EAGERx.',
    'long_description': None,
    'author': 'Jelle Luijkx',
    'author_email': 'j.d.luijkx@tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eager-dev/eagerx_gui',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
